% Implemented by Henning Christiansen 1995 for the Demo system
%   
%
% Adapted for the new DemoII version of the system 1998 by
% Davide Martinenghi
%
% See legal notice in the end of this file.
%
/*
Load file as follows:

['Examples/DefaultLogic'].

*/

%%%% An extension of example 3.1 in the Kakas/Kowalski/Toni paper
%
% Implemented by Henning Christiansen 1995, updated 1996
% AND 1998
% using the Demo-system ---
% Contact: henning@dat.ruc.dk

% PLAYING AND EXTENDING THE EXAMPLE:
%   - default logic by generating instances of default RULEs
%   - ABDUCTION new extensional facts under default rules
%   - INDUCTION of new default-and-exceptions


%%%%%%%%%%%%%%%%%%%%%%
%
% First excercise: A theory with a fixed default rule

% 'Background knowledge' (but not the default rule) given by:

object_module( kb0,

     \[ (bird('X',yes):- penguin('X',yes)),
        (fly('X', no):- penguin('X',yes)),
         penguin(tweety, yes),
         bird(john, yes),
         bird(norma, yes)
      ]).

% A default rule such as
%
%    fly(X, yes):- bird(X, yes)
%
% is not applied directly, only instances of it,
% Sequences of such instances are defined as follows:

:- block default_instances(?,-).

default_instances(_,\[]).

default_instances(\ (?P^[?_, yes]:- ?Q^[?_, yes]),
                  \ [(?P^[?C, yes]:- ?Q^[?C, yes] ) | ?Ds] ):-
    constant_(C),
    default_instances(\ (?P^[?_, yes]:- ?Q^[?_, yes]) ,Ds).

% Provability in a default logic where birds normally fly
% is defined as follows.

demo_birds_fly(KB, Q, DIs):-
    default_instances( \ (fly('X', yes):- bird('X', yes)), DIs),
    demo( \ ( ?KB & ?DIs), Q),
    close_constraints( \ ( ?KB & ?DIs) ),
    consistent( \ ( ?KB & ?DIs) ).

% (NB: The predicates return the instances used of the default rule,
%  unnecessary, but nice to look at )


consistent(T):-
    fails( T, \\ ( 'P' ^['X', yes], 'P' ^['X', no])).

/*
 Test runs

%   yes
demo_birds_fly( \kb0, \\fly(john,yes), _ ).

%    no
demo_birds_fly( \kb0, \\fly(tweety,yes), _ ).

*/

%%%%%%%%%%%%%%%%%%%%%%
%
% Second excercise: Extend with abduction to update extensionals

% ... so we must formalize what 'extensionals' means:

:- block extensionals(-).

extensionals( \ []).

extensionals( \ [ (bird(?X, yes):- true) | ?More ]):-
    constant_(X),
    extensionals( More ).

extensionals( \ [ (penguin(?X, yes):- true) | ?More ]):-
    constant_(X),
    extensionals( More ).

% Update by abduction is now straightforward:

update(KB, Observation, NewKB, D):-
    extensionals(UpdateFacts),
    demo_birds_fly( \ (?KB & ?UpdateFacts), Observation, D),
    NewKB = \ ( ?KB & ?UpdateFacts ).

/*
 Test runs

update( \kb0, \\fly(peter, yes), N, D).
    
 D = \ (fly(peter,yes):-bird(peter,yes))
 N = \ (kb0 & [(bird(peter,yes):-true)]) ? ;

 no
*/

%%%%%%%%%%%%%%%%%%%%%%
%
% Third excercise: Induction of defaults-with-exceptions
%                  from examples....

% .. so we have to formalize what we accept as a default rule...

default_rule(\ (?P^[ 'X', yes]:- ?Q^[ 'X', yes] )):-
    dif(P,Q).


% ... and formalize what we accept as exceptions to a given
%     default rule....

% exception_rules( default-rule, program-of-exceptions-rules)

:- block exception_rules(?,-).

exception_rules(_, \[]).

exception_rules(Default, \[?ExR | ?ExRs]):-
    exception_rule(Default, ExR),
    exception_rules(Default, ExRs).

exception_rule(\ (?P ^ ?_ :- ?Q ^ ?_),
               \ (?P^[ 'X', no]:- ?Ex^[ 'X', yes] )):-
    dif(Ex, P), dif(Q,Ex).

   % ... one learns by experience that the dif stuff
   %     needs to be there!


% ... finally, put the whole thing together around a call to demo:

aha(Facts, Obs, Def, Excs):-
    default_rule(Def),
    default_instances(Def, DefIs),
    exception_rules(Def, Excs),
    demo(\ (?Facts & ?Excs & ?DefIs), Obs),
    close_constraints(\ (?Facts & ?Excs & ?DefIs) ),
    consistent(\ (?Facts & ?Excs & ?DefIs) ).


% We use the following sample 'background knowledge':


object_module( facts,

     \[ (bird('X',yes):- penguin('X',yes)),
        (bird('X',yes):- chicken('X',yes)),
         penguin(tweety, yes),
         bird(john, yes),
         bird(peter,yes),
         chicken(andrew,yes)
      ]).


% Try the following to test our little induction machine:

/*
testaha1(D,E).

D = \ (fly('X',yes):-bird('X',yes)),
E = \[(fly('X',no):-penguin('X',yes)),(fly('X',no):-chicken('X',yes))] ? ;

no

testaha2(D,E).

D = \ (fly('X',yes):-bird('X',yes)),
E = \[(fly('X',no):-penguin('X',yes))] ? ;

no
*/

testaha1(D,E):-
    aha(\facts, \\ (fly(john,yes), fly(peter,yes), fly(tweety,no),
                    fly(andrew,no)),
    D,E).

testaha2(D,E):-
    aha(\facts, \\ (fly(john,yes), fly(peter,yes), fly(tweety,no),
                    fly(andrew,yes)),
    D,E).

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
