% Implemented by Henning Christiansen 1995 for the Demo system
%   
%
% Adapted for the new DemoII version of the system 1998 by
% Davide Martinenghi
%
% See legal notice in the end of this file.
%
% Load file as follows:
/*
   ['Examples/UpdateFamily'].
*/

object_module( kb, \ [
		(sibling('X','Y'):- dif('X','Y'), parent('Z','X'), parent('Z','Y')),
		(parent('X','Y'):- father('X','Y')),
		(parent('X','Y'):- mother('X','Y')),
		(father(john, mary):- true),
		(mother(jane,mary):- true)
		]).

object_module( ic, \ [
		(bottom:- dif('A','B'),father('A','C'), father('B','C')),
		(bottom:- dif('A','B'),mother('A','C'), mother('B','C')),
		(bottom:- father('A','Z'), mother('A','X'))	% hermaphrodite test
		]).

object_module( ic2, \ [ %% no hermaphrodite test here
		(bottom:- dif('A','B'),father('A','C'), father('B','C')),
		(bottom:- dif('A','B'),mother('A','C'), mother('B','C'))
		]).

update(ObsFacts, UpdateFacts):-
    extensionals(UpdateFacts),
    fails( \ (kb & ic & ?UpdateFacts), \\bottom ),
    NewKB= \ (kb & ?UpdateFacts),demo( NewKB, ObsFacts ),
    close_constraints( NewKB ).

update2(ObsFacts, UpdateFacts):- %% without hermaphrodite test
    extensionals(UpdateFacts),
    fails( \ (kb & ic2 & ?UpdateFacts), \\bottom ),
    NewKB= \ (kb & ?UpdateFacts),demo( NewKB, ObsFacts ),
    close_constraints( NewKB ).

:- block extensionals(-).

extensionals( \ []).

extensionals( \ [ (father(?X, ?Y):- true) | ?More ]):-
    constant_(X), constant_(Y),
    extensionals( More ).

extensionals( \ [ (mother(?X, ?Y):- true) | ?More ]):-
    constant_(X), constant_(Y),
    extensionals( More ).

/*
  Execution examples

| ?- update( \\sibling(mary,bob), N).

N = \[(father(john,bob):-true)] ? ;
N = \[(mother(jane,bob):-true)] ? ;
no

| ?- update(\\ (sibling(mary,bob),mother(joan,bob)),N).

N = \[(father(john,bob):-true),(mother(joan,bob):-true)] ? ;
no

%% This run is made without the hermaphrodite test:
| ?- update2( \\sibling(mary,bob), N).

N = \[(father(john,bob):-true)] ? ;
N = \[(mother(john,bob):-true)] ? ;
N = \[(father(jane,bob):-true)] ? ;
N = \[(mother(jane,bob):-true)] ? ;
no
*/

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
