% Implemented by Davide Martinenghi 1999 for the Demo system
%
% See legal notice in the end of this file.
%
% Load file as follows:
/*
	['eventcalculus/ec'].
*/

:-use_module(library(lists)).

%%% SCHEMATIZATION OF EVENT CALCULUS
% as in paper "Reinventing Shakey", Shanahan, M. 1999, Tracking Number: A319, to appear.
% but with instantaneous actions (happens/2 instead of happens/3)
object_module( ec,
     \[
/*EC1*/	(holdsAt(yes,'F','T') :- initiallyTrue('F'), clipped(no,t0,'F','T')),

/*EC2*/	(holdsAt(yes,'F','T2') :- happens('A','T1'), <('T1','T2'), initiates('A','F','T1'),
     		clipped(no,'T1','F','T2')),

/*EC3*/	(clipped(yes,'T1','F','T2') :- happens('A','T'), <=('T1','T'), <('T','T2'),
     		terminates('A','F','T')),

/*EC4*/	(holdsAt(no,'F','T') :- initiallyFalse('F'), declipped(no,t0,'F','T')),

/*EC5*/	(holdsAt(no,'F','T2') :- happens('A','T1'), <('T1','T2'), terminates('A','F','T1'),
     		declipped(no,'T1','F','T2')),

/*EC6*/	(declipped(yes,'T1','F','T2') :- happens('A','T'), <=('T1','T'), <('T','T2'),
     		initiates('A','F','T'))
        ]).


%%% DEFINITION OF THE PARTIAL ORDER FOR THE TIMEPOINTS
object_module( less_or_equal,
     \[ (<=('A','A')),
        (<=('A','B') :- <('A','B'))
	%(<('A','C') :- <('A','B'),<('B','C'))
	]).

%%% INTEGRITY CONSTRAINTS

% PRINCIPLE OF NON-CONTRADICTION
object_module(nonContradiction,
     \[
	(bottom :- clipped(no,'A','B','C'),clipped(yes,'A','B','C')),
        (bottom :- declipped(no,'A','B','C'),declipped(yes,'A','B','C'))%,
%        (bottom :- holdsAt(yes,'A','B'),holdsAt(no,'A','B'))
      ]).

% CONSISTENCY OF THE PARTIAL ORDER
object_module(timeConsistency,
     \[
        (bottom :- <('T','T')),
	(bottom :- <('_',t0))
      ]).


%%% Pattern for the plan
easy_mk_bias_predicate( 
    plan, 
    \ happens(?A, ?T), 
    (action(A), timepoint(T))).
    
%%% Pattern for abduced negative conditions
easy_mk_bias_predicate( 
    negative_conditions, 
    \ (?P ^ [no, ?T1, ?F, ?T2]), 
    ((P=clipped ; P=declipped), timepoint(T1), fluent(F), timepoint(T2))).

%%% Pattern for abduced relations between timepoints
easy_mk_bias_predicate( 
    timepoints, 
    \ <(?T1, ?T2),
    (timepoint(T1), timepoint(T2),ground_(\\ <(?T1,?T2)))).


lightec(Domain,Goal,\ (?P & ?N & ?T)) :-
	preds_(P,[happens]), plan(P),
	preds_(N,[clipped,declipped]), negative_conditions(N),
	preds_(T,['<']), timepoints(T),
	fails(\ (?Domain & ?T & timeConsistency), \\ bottom),
	demo(\ (?Domain & ec & less_or_equal & ?P & ?N & ?T), Goal),
	close_constraints(\ (?P & ?N & ?T)).

ec(Domain,Goal,\ (?P & ?N & ?T)) :-
	preds_(P,[happens]), plan(P),
	preds_(N,[clipped,declipped]), negative_conditions(N),
	preds_(T,['<']), timepoints(T),
	fails(\ (?Domain & ?T & timeConsistency), \\ bottom),
	fails(\ (?Domain & ec & less_or_equal & nonContradiction & ?P & ?N & ?T),\\bottom),
	demo(\ (?Domain & ec & less_or_equal & ?P & ?N & ?T), Goal),
	close_constraints(\ (?P & ?N & ?T)).

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/demo.html
