% Implemented by Davide Martinenghi 1999 for the Demo system
%
% See legal notice in the end of this file.
%
% Make sure the event calculus main program 'eventcalculus/ec.pl' is loaded
% and load the file as follows:
/*
	['eventcalculus/key'].
*/


% Domain taken from paragraph 2.1 of the following paper:
% Miller, R., and Pinto, J. "Temporal Languages for Reasoning About Action", November 7, 1998.
object_module( domain,

     \[
	%%% DOMAIN
     	initiates(pickup,haskey,'T'),
        (initiates(insert,locked,'T') :- holdsAt(no,locked,'T'),holdsAt(yes,haskey,'T')),
        (initiates(gothrough,inside,'T') :- holdsAt(no,locked,'T'),holdsAt(no,inside,'T')),

        (terminates(gothrough,inside,'T') :- holdsAt(no,locked,'T'),holdsAt(yes,inside,'T')),
        (terminates(insert,locked,'T') :- holdsAt(yes,locked,'T'),holdsAt(yes,haskey,'T')),
        
	%%% NARRATIVE
	initiallyTrue(locked),
	initiallyTrue(inside),
	initiallyFalse(haskey),

	happens(pickup,t2),
	happens(insert,t4),
	happens(gothrough,t6),

	%%% TIMEPOINT ORDERING
	t0<t2,t0<t4,t0<t6,t0<t8,
	t2<t4,t2<t6,t2<t8,
	t4<t6,t4<t8,
	t6<t8
	]).

%%% Definition of the sorts of the language:
fluent(\\\ haskey). % the robot has the key
fluent(\\\ locked). % the door is locked
fluent(\\\ inside). % the robot is inside
action(\\\ pickup). % the robot picks up the key
action(\\\ insert). % the robot inserts the key in the door
action(\\\ gothrough). % the robot goes through the door
timepoint(T) :- constant_(T).

%%% GOALS:
/*
 ec(\domain,\\holdsAt(no,inside,t8),A).

A = \ (	[] & % plan

	[(clipped(no,t0,locked,t4):-true),
	(clipped(no,t2,haskey,t4):-true),
	(declipped(no,t4,locked,t6):-true),
	(clipped(no,t0,inside,t6):-true),
	(declipped(no,t6,inside,t8):-true)] &

	[]) % timepoints

do not ask for alternative solutions!!!!
*/

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/demo.html
