% Example described in section 4.2, p. 242 ff of the paper
%
% Christiansen, H.,
% Automated reasoning with a constraint-based metainterpreter,
% Journal of Logic Programming, 1998. Vol 37(1-3) Oct-Dec 98, pp. 213-253.
% 
% Implemented by Henning Christiansen 1995 for the Demo system
%   
%
% Adapted for the new DemoII version of the system 1998 by
% Davide Martinenghi
%
% See legal notice in the end of this file.
%
% Load file as follows:
/*
   ['Examples/wet_grass'].
*/

% SIMPLE ABDUCTION
% ================
%
% Produced by Henning Christiansen 1994
%
% Here we show an extremely simple example of abduction
% which appears often in the general literature on
% abduction.

object_module( garden,

	\[ (grass_is_wet:- rained_last_night),
	(grass_is_wet:- sprinkler_was_on)
	]).

% Sample query:
%
/*
	demo( \(garden & [?WHY]), \\grass_is_wet ).
*/
%
%  --- First solution: WHY = \(rained_last_night:- true)
%  --- More solutions?
%      LOOP since demo now explores the possibility
%
%           \(rained_last_night:- ?ATOM)
%
%      there atom should be proved using a clause in the program,
%      and (alas, depth-first) it selects the first one, i.e.,
%      ATOM is at atom which unifies with 'rained_last_night'
%      .....................

% Lets refine...

:- block abducible(-).

abducible( \ (rained_last_night:- true) ).
abducible( \ (sprinkler_was_on:- true) ).
abducible( \ (full_moon:- true) ).
abducible( \ (sun_shines:- true) ).
abducible( \ (dog_on_lawn:- true) ).

% Sample query:
%
/*
	abducible(WHY), demo( \(garden & [?WHY]), \\grass_is_wet ).
*/
%
% --- Now we get the desired behaviour...

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
