%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: close_constraints.pl

% and related stuff, incorporating some lgg like stuff.


% Program tails always instantiated to empty

% Metavariables of type T instantiated as follows:
%
%   no-type     ->  '*weird*'
%
%   program     ->   \[]
%
%   clause      ->  \ (<dummy-pred1> :- <dummy-pred2>)
%
%   formula     ->  \ <dummy-pred>  (?)
%
%   atom        ->  \ <dummy-pred>  (?)
%
%   constraint  ->  <dummy-const> = <dummy-const>
%                   or dif(<dummy-const1>, <dummy-const2>)
%
%   term
%   1) with instance constraints waiting on it
%  
%               ->  take all instance constraints
%                   frozen on the variable and
%                   run an lgg on them
%                   to get some term name.
%                   (( PERHAPS THE TESTS FOR SHARING TAKES
%                       CARE OF RANGE-RESTRICTED CASES?))
%                !! In case some constraint or side-condition
%                   provokes failure, lift gradually to more general terms
%
%   2) no instance constraints
%
%               ->  if not covered by any type constraint,
%                   generate a new constant.
%                   if covered by syntax constraints,
%                   generate a new thing of that kind.
%               
%  If any of the above do does not work, 
%  do not instantiate and issue a warning.


%% THE FOLLOWING PREDICATE USES ASUMPTIONS
%% ABOUT WHICH TYPES THAT HAVE BEEN DECLARED
%% AND ALSO ABOUT WHICH SORTS OF CONSTRAINTS
%% THAT CAN (AND CANNOT) BE IN A REDUCED STATE

%% NB: A new Heuristics: We TRY to instantiate to our
%%     most qualified guess, and if failure
%%     occurs, well, then do not instantiate
%% DO THIS WHEN IT HAS BEEN DEBUGGED.

:- use_module(library(terms)).

close_constraints(X):-
   reset_object_thing_generator_counters,
   close_constraints_bt(X), !,
   (ground(X) -> true
   ;
   print_message(warning,demo(close_constraints(X))),
   fail, %%%% TEMPORARY!!!!!!!!
   nl).

% the suffix '_bt' stands for backtrack, - internally
% close_constraints can backtrack a little, but
% in total, only produce one solution

close_constraints_bt(X):-
    nonvar(X), !,
    (X = [] -> true
     ;
     X = [H|T] -> close_constraints_bt(H), close_constraints_bt(T)
     ;
     X =.. [_|Xs],
     close_constraints_bt(Xs)).




close_constraints_bt(Term):-  % a variable 

    type_constrained(Term,program)
    -> ( Term = \[]
         ;
         % IN CASE SIDE-CONDITIONS => FAILURE, DO NOT INSTANTIATE
         true
       )
    ;
    type_constrained(Term,clause)
    % Assumption: No non-trivial instance constraints in it
    -> ( new_object_predicate(P1),
         new_object_predicate(P2),
         Term =  \ (?P1^[] :- ?P2^[])
         ;
         % IN CASE SIDE-CONDITIONS => FAILURE, DO NOT INSTANTIATE
         true
       )
    ;
    type_constrained(Term,atom)
    % Assumption: No non-trivial instance constraints in it
    -> ( new_object_predicate(P),
         Term =  \\ (?P^[])
         ;
         % IN CASE SIDE-CONDITIONS => FAILURE, DO NOT INSTANTIATE
         true
       )
    ;
    type_constrained(Term,constraint)
    % Assumption: No non-trivial instance constraints in it
    -> ( new_object_constant(C1),
         % two options - just in case of stupid user constraints
         (Term = \\ (?gr(constant,C1,[]) = ?gr(constant,C1,[]))
          ;
          new_object_constant(C2),
          Term = \\ dif(?gr(constant,C1,[]),?gr(constant,C2,[])))
          ;
          % IN CASE SIDE-CONDITIONS => FAILURE, DO NOT INSTANTIATE
          true
       )
    ;
    type_constrained(Term,term)
    -> ( chr:findall_constraints( Term,
              instance_(Term,_,_), InstList),
         InstList \== []

         -> ( same_top_fun(InstList, Fun, Arity),
              \+ functor(Fun,metavar,1)
              -> (Arity == 0
                  -> ( Term = gr(constant, Fun,[])
                       % !!! If this fails lift up to slightly more general
                       ;
                       (var_to_share_with(InstList, ShareVarId)
                         -> Term = gr(variable, ShareVarId, [])
                          ;  new_object_variable(Id),
                           Term = gr(variable,Id,[])
                       )
                       ;
                       % !!!If this also fails, do not instantiate
                       true
                     )
                   ;
                    length(ArgList, Arity),
                    ( Term = gr(structure, Fun, ArgList),
                      close_constraints_bt(ArgList) %%%-> true 
                      % !!! If this fails lift up to slightly more general
                      ;
                      (var_to_share_with(InstList, ShareVarId)
                        -> Term = gr(variable, ShareVarId, [])
                         ;  new_object_variable(Id),
                          Term = gr(variable,Id,[])
                      )                      ;
                       % !!!If this also fails, do not instantiate
                       true
                    )

                  )
               ;
               % No common structure, hence a variable:
               %   (NB; takes care of two cases: different fun's and same metavar!)
               % Should variable be reused due to sharing?
               var_to_share_with(InstList, ShareVarId)
               -> Term = gr(variable, ShareVarId, [])
               ;
               new_object_variable(Id),
               Term = gr(variable,Id,[])   % CONTEXT??
             )
          ;
          % TERM WITH NO INSTANCE CONSTRAINTS
          ( generate_dummy_ground_term(Term) %%% -> true
            ;
            % IN CASE SIDE-CONDITIONS => FAILURE, DO NOT INSTANTIATE
            true
          )
       )
       ; (Term = '*weird*' ; true).


generate_dummy_ground_term(X):-
   ( type_constrained(X, variable)
     -> new_object_variable(V), NewX = gr(variable,V,[])
     ;
     type_constrained(X, structure)
     -> new_object_function(F), new_object_constant(C),
        NewX = gr(function, F,[gr(constant,C,[])]),
       X = NewX
    ;
    new_object_constant(C), NewX = gr(constant,C,[])),
  X = NewX.
    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% same_top_fun(In: InstanceList, CommonFunctor, Arity)
%
% InstanceList is in the format provided by findall_constraints
%
% Succeeds if either
% - First arg. is a list of identical metavariables
%   in which case CommonFunctor = metavar(<that variable>).
%   In this case Arity be ignored
% or
% - All elem's in first list are names of
%   terms with primary CommonFunctor/Arity.



same_top_fun([],_,Arity):-
    nonvar(Arity) -> true ; Arity=0.

same_top_fun([ #(instance_(_,T1,_),_) |Ts], Functor, _):-
    var(T1), % special case, viewed as a constant in some sense
    !,
    (var(Functor) % means first iteration
     -> Functor = metavar(T1)
     ; Functor == metavar(T1)), % succeed only if all guys have
                                % same metavar
    same_top_fun(Ts, Functor, _).

same_top_fun([#(instance_(_, gr(structure,Functor,Args), _),_)|Ts],
             Functor, Arity):-

    when( (nonvar(Arity); nonvar(Args)),
          length(Args,Arity)),
    same_top_fun(Ts, Functor, Arity).


same_top_fun([#(instance_(_, gr(constant,Functor,[]), _),_)|Ts],
          Functor, 0):-
    same_top_fun(Ts, Functor, 0).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% var_to_share_with(In: InstanceList, VArName)
%
% InstanceList is in the format provided by findall_constraints
%
%

var_to_share_with([],_).

var_to_share_with([ #(instance_(_,T,Sub),_) |Ts], ShareVar):-
    is_member( (ShareVar -> ShareT), Sub),
    T == ShareT,
    var_to_share_with(Ts, ShareVar).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliaries

is_member(_,L):- var(L),!,fail.

is_member(M,[M|_]).

is_member(M,[_|L]):- is_member(M,L).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type_constrained(Object, +Type)
%
% - checks if Object at the top level looks like
%   an object phrase of type Type.
% - if Object is given as a variable, we take
%   a look at the possible type or instance constraints
%   concerning Object.
%
% (this way, we chack without setting up new constraints)

% Restrictions: Object should be 'external'
% and Type must be given

type_constrained(X, Type):-
    var(X) ->
        ( chr:find_constraint( X, type(ObsType, X),_)
          -> ( Type=ObsType ; subtype(Type, ObsType))
         )
     ; X = gr(ObsType,_,_),
       ( Type=ObsType ; subtype(Type, ObsType)).


% DELETE?
% object_thing_in_object_phrase(Kind, ReturnThing, ObjectAny)
% Goes through ObjectAny and returns a thing of type Kind
% used in it.

object_thing_in_object_phrase(_,_,Var):- var(Var), !, fail.

object_thing_in_object_phrase(Kind,gr(Kind,Id,L),
                                   gr(Kind,Id,L)):- !.

object_thing_in_object_phrase(V, gr(_,_,L)):-
    nonvar(L),
    member(T,L),
    nonvar(T),
    object_thing_in_object_phrase(V,T).


% Generating new objects


new_object_variable(OV):-
    next_object_thing_id(variable, OV).

new_object_constant(OV):-
    next_object_thing_id(constant, OV).

new_object_function(OV):-
    next_object_thing_id(function, OV).

new_object_predicate(OV):-
    next_object_thing_id(predicate, OV).


%%%% Counters:

:- dynamic object_thing_no/2.

object_thing_no(variable, 0).
object_thing_no(constant, 0).
object_thing_no(function, 0).
object_thing_no(predicate,0).


reset_object_thing_generator_counters:-
   retractall(object_thing_no(_,_)),
   asserta(object_thing_no(variable, 0)),
   asserta(object_thing_no(constant, 0)),
   asserta(object_thing_no(function, 0)),
   asserta(object_thing_no(predicate, 0)).

next_object_thing_id(Kind,Symbol):-
    (Kind=variable -> Prefix = "NewVar"
     ; Kind=constant -> Prefix = "newConst"
     ; Kind=function -> Prefix = "newFun"
     ; Kind=predicate -> Prefix = "newPred"),
   retract(object_thing_no(Kind, VN)),
   VN1 is VN+1,
   asserta(object_thing_no(Kind, VN1)),
   number_chars(VN, NumberChars),
   append(Prefix, NumberChars, IdentString),
   name(Symbol, IdentString).
  
%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
