%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: df.pl

%  fails(Program-name, Query-name)
%
%  A lazy version of negation-as-failure which delays subcomputations
%  that depend on missing components of the object program.
%  - it provides an incremental evaluation of integrity constraints
%    and similar things.

%   See discussion about the overall principles in the end of this file.
% THE MENTIONED COMMENTS ARE OUT OF DATE
% - THIS AN EXPERIMENT BY ADDING AN EXTRA ARGUMENT FOR KEEPING
%   TRACK OF WHEN DEAD CONSTRAINTS CAN BE CLEANED UP.
%   THIS IS NOT EXPLAINED IN THE FILE YET
%   BUT ARGUMENTS CALLED Nilfisk (WITH POSSIBLE SUBSCRIPTS)
%   TAKE CARE OF THIS MACHINERY

% NOTICE: We have given priority to readable answers rather than ultimate efficiency
%         (the device is too slow anyhow)

:- demo_use_module('Source/df_uti').

constraints fails1/3,
            fails1_with_all_clauses/5,
            fails1_with_clause/5,
	    delay_remove_dead_constraints/2.

% fails(Program-name, Query-name)

fails(P, Q):-
    instance_(Q, Q1),
    delay_remove_dead_constraints(Nilfisk,Q1),
    fails1(P, Q1, Nilfisk).

% Remark: The variable name "Nilfisk" is taken from a Danish brand of vacuum cleaners

% fails1(Program-name, internal-Query-name, Nilfisk)
%

%%% Rule F1
fails1(_, \\true, _) <=> fail.


%%% Rule F2
fails1(_, \\ (?A = ?B), Nilfisk) <=> ?=(A,B), A\==B |
	Nilfisk=g.

%%% Rule F3
fails1(_, \\ (?A = ?B), _) <=> unifiable_independently_of_external_variables(A,B) |
	fail.


%%% Rules F5, F6, F7 & F8
fails1(P, Q, Nilfisk) <=>
      can_select_simple(Q, Simple, Cont)
   |
   
      (var(Simple) 
       -> % by def of can_select_simple, Simple is constrained to be an atom
          fails1_with_all_clauses(P, Simple, P, Cont, Nilfisk)
       ;
       Simple = \\true
       -> fails1(P, Cont, Nilfisk)
       ;
       Simple = gr(atom,_,_)
       ->  fails1_with_all_clauses(P, Simple, P, Cont, Nilfisk)
       ;
       Simple = \\ (?A = ?B)
       -> % by def of can_select_simple, Simple does not depend on external stuff
          (A=B
           -> fails1(P, Cont, Nilfisk)
           ;  Nilfisk = g)
       ;
       Simple = \\ dif(?A, ?B)
       -> % by def of can_select_simple, ?=(A,B)
          (A==B
           -> Nilfisk = g
           ; fails1(P, Cont, Nilfisk))
          
       ;
       Simple = \\true % probably never happens?
       -> fails1(P, Cont, Nilfisk)
       ;
       Simple = \\ (?A ; ?B )
       -> fails1_disjunction(P, A, B, Cont, Nilfisk)
      ).


%%% Rule F4
% Easier cases are taken care of by rules F5-F8
% If there is a bunch of difs (and nothing else),
% we choose them and process them one at a time on backtracking
% (and leave the other ones unprocessed)
fails1(_, Difs, Nilfisk) <=> only_dif_constraints(Difs) |
   select_dif(Difs,\\dif(?A,?B)),
   skolemize_internal_variables( \\dif(?A,?B) ),
   A=B, Nilfisk = g.




% fails1_with_all_clauses(Clauses, Atom, Prog, Continuation, Nilfisk)
fails1_with_all_clauses(\ [?C | ?Cs], A, P, Cont, Nilfisk) <=>
    Nilfisk = (Nilfisk1,Nilfisk2),
    fails1_with_clause(C, A, P, Cont, Nilfisk1),
    fails1_with_all_clauses(Cs, A, P, Cont, Nilfisk2).

fails1_with_all_clauses(\ (?C1 & ?C2), A, P, Cont, Nilfisk) <=>
    Nilfisk = (Nilfisk1,Nilfisk2),
    fails1_with_all_clauses(C1, A, P, Cont, Nilfisk1),
    fails1_with_all_clauses(C2, A, P, Cont, Nilfisk2).

fails1_with_all_clauses(Cs, A, _, _, Nilfisk) <=>
     var(Cs), incompatible_preds_(Cs, A)
    |
     Nilfisk = g.

fails1_with_all_clauses(\[], _, _, _, Nilfisk) <=> Nilfisk = g.

%%%%%%%%%%%%%%%%%%%%%%%%
%
% fails1_with_clause(C, A, P, Cont, Nilfisk)
% - C is a clause (external), i.e., needs to be processed by
%   instance_ before it can take part in interesting unifications
% - A is an atom (internal)
% - P is the whole program (external) as usual
% - Cont is the current rest-of-goals (internal)
%



% This a preliminary test: this rule checks preds_ when the clause is a variable
fails1_with_clause(C, A, _, _, Nilfisk) <=> var(C), incompatible_preds_(C, A) |
	Nilfisk=g.



% Case 1. A simple analysis (without involving instance_) says
%         that unification can never succeed.
%         - NB: this one catches also some cases where case 2 below
%           could have been applied with unification failing.
%         - However, the rule for case 1 is much more efficient,
%           so that is why it is no. *1*

% Notice: the following applies even when Selected Atom is a variable
%         covered by atom_( ).

% A first rule - should be polished!
fails1_with_clause(\ (?Head :- ?_), SelectedAtom, _, Cont, Nilfisk)  <=>

    atoms_cannot_instance_(Head, SelectedAtom, Cont) % Continuation is looked
                                                     % at for possible dif's

  |
    Nilfisk=g.


% Case 2. There is enough information present in order to
%         take instance_ of head-of-clause and perform
%         a unification whose result can be trusted.
fails1_with_clause(\ (?Head :- ?Body), SelectedAtom, Prog, Cont, Nilfisk) <=>

    (  (ground(Head),
        no_more_changes_implied_by_meta_level(SelectedAtom))
    ; 
       (subsumes_chk1(SelectedAtom, Head),
        no_more_changes_implied_by_meta_level(SelectedAtom))
    )       

   |
          
     ( % Unfolding instance_ and mk_internal_variant for efficiency

       (mk_internal_variant(SelectedAtom,SelectedAtomCopy,V,W),
        instance_(Head, SelectedAtomCopy, S) % embeds unification
        -> (instance_(Body, Body1, S), 
            mk_internal_variant(Cont,ContCopy,V,W),
            conjunc(Body1, ContCopy, NewQuery),

            delay_remove_dead_constraints(Nilfisk, \\ (?SelectedAtomCopy, ?NewQuery)),

            fails1(Prog, NewQuery, Nilfisk))
           
         ; Nilfisk=g,
             remove_dead_constraints(SelectedAtomCopy)
        )
      ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% fails1_disjunction(P, A, B, Rest, Nilfisk)
%

fails1_disjunction(P, A, B, Rest, (Nilfisk1,Nilfisk2)):-
    mk_internal_variant(\\ (?B, ?Rest), BRestCopy,_,_),
    fails1(P, \\ (?A, ?Rest), Nilfisk1),
    delay_remove_dead_constraints( Nilfisk2, BRestCopy),
    fails1(P, BRestCopy, Nilfisk2).


%%%%%%%%%%%%%%
%
% delay_remove_dead_constraints( Nilfisk, Query)
%
/* old implementation with 'when'
delay_remove_dead_constraints( Nilfisk, Query):-
    ground(Query) -> true
   ;
    when( ground(Nilfisk), remove_dead_constraints(Query)).
now it is implemented as a constraint (slower but more elegant and controllable)
*/
delay_remove_dead_constraints(_, Query) <=> ground(Query) |
	true.

delay_remove_dead_constraints(Nilfisk, Query) <=> ground(Nilfisk) |
	remove_dead_constraints(Query).

/* testdata

fails(\[(a)],\\a).

fails(\[(b)],\\a).

fails(\[(?H :- true)],\\f(?A)), H= \\b.

fails(\[(?H :- true)],\\f(c,?A,a)), H= \\f(d,?B,b).

fails(\[(?H :- true)],\\f(a)), H= \\ ('P'^['X']).

The use of subsumes_chk in last rule above:

fails( \ [ p(?A,?A), (bot:- p('X','X'))], \\bot).
*/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Discussion of the method used in fails
%
% 1. A simplified presentation of demo_fail's implementation method
% =================================================================
% 
% The constraint handling rules that implement fails look a bit
% complicated due to the explicit copying of terms and constraints
% and the following removal of these constraints.
%
% However, at the conceptual level, it can be described in a much
% simpler way without all this machinery - so it becomes meaningful
% to describe it in a publication and to consider proofs of soundness.
%
% The trick is the following: 
% - Instead of *one* call to fails, we send off
%   *sufficiently many* identical calls of it.
% - Replace rules for fails1_with_all_clauses by one which
%   selects nondeterministically one of the clauses or decides to
%   delay on the tail.
% - Assume this nondetermininism is fair!
% 
% Each execution of one of the many calls to fails corresponds to
% one "path" in an SLD tree - and each of those paths must lead to an
% object level faliure in order to have the the whole stuff to
% fail at object level, i.e., to have "failure" to succeed at the meta_level.
%
% And to have the multitude of fails calls to execute in the same
% state is equivalent to say that all of the must succeed in order
% to get the 'overall' metalevel success.
%
% In this way, the rules becomes few and can be understod in a
% fully declarative way. And the copy stuff can be explained by a bit
% of armwaiving as a way to implement thes use of nondeterminism!
%
%
% 2. Replacing nondeterminism by copying of terms and constraints
% ===============================================================
%
% The nondeterminisms mentioned above needs to be replaced by something
% more operational - and instead of selection a single clause
% in fails1_with_all_clauses, we so to speak select all of them in parallel
% and give them each a fresh copy of the (the relevant portion of) the
% execution state. 
%
% 2.1 A conceptually simple but inefficient medthod
% -------------------------------------------------
%
% This is a first step to bridge the method described in section 1 to
% the actual implementation contained in this file.
%
% We could have the constraint handling rules keep a trace over all
% those instance_ constraints and unifications that has been used in
% order to reach each point in the computation. Each time
% fails1_with_all_clauses (now not nondeterministic) splits up
% for each clause in the program, we execute the mentioned trace
% (with new variables as substitution arg's to instance_) to
% produce relevant arguments to the following computations
%
% This can be described as a semantics-preserving program transformation
% from the nondeterministiv version.
%
% However, this would add a quadratic factor or so in the overall computation,
% - and a serous one, because evaluation of instance_ constraints seems to be
% *the* critical of the whole system.
%
% 2.2 Copying object query and attached constraints
% -------------------------------------------------
%
% An easier way to acheive the same result as produced by the inefficient
% method described in section 2.1, is explicitly to produce a copy of
% the current query (=selected atom + rest-of-goals) - one for each
% clause to be investigated.
%
% We use a 'copy machine' that replaces some variables consistently
% - those 'internally' generated - but keeps those that are constrained
% to be meta_level_constrained.
% 
% If an internal variable X is constrained by instance_(..., ...X..., ...)
% we need to produce a copy of that constraint - but keeping the first
% (and 'external') argument of it in the original form.
%
% In this way the copy of the query reacts exactly in the same way
% to external events as does the query from which it is copied.
% 
% We have to take care in order to copy all relevant instance constraints
% as described above. Consider for example a constraint
%
%    instance(..., ...X..., ....)
%
% or  
%
%    instance(...,   .... , [ ... (..., ...X...) ]).
% 
% where X appears in the query Q being copied. 
%
% Assume now, that some internal variable Y which is *not* in Q
% but which occurs in an instance constraint pending
% on the above mentioned X. This might be as follows:
%
%    C_X = instance(..., ..\\\f(?X,?Y)..., ....)
%
% and assume there is another instance constraint
%
%    C_Y = instance(..., ...Y..., ....)
%
% where this constraint does not contain a variable of Q.
%
% In this case, we produce a copy of C_X,
%
%    C_X' = instance(..., ...\\\f(?X',?Y')..., ....)
%
% as well as a copy of C_Y.
%
%    C_Y' = instance(..., ...Y'..., ....).
%
% Only in this way, we can be sure to maintain a all effects
% of external events from the original query.
% As an example, assume that the first argument of C_Y (identical to
% first arg. of C_Y') acheives a value which causes C_Y to instantiate
% Y to, say, \\\a, then C_Y' causes Y'= \\\a. 
% This change propagation is crucial if, for example, \\\f(?X',?Y') is
% unified in the process with the term \\\f(?Z,?Z).
%
% To implement this, we evaluate a fixpoint throughout all those constraints
% that are pending on those variables that we have seen in the constraints
% we have considered. However, in most cases one or two iterations are 
% sufficient. 
%
% 2.3 Removing 'dead' constraints
% -------------------------------
% 
% The constraint solver should make sure that each branch is execution
% on its own (and 'correct') copy of the current query.
% If, now, object-level failure is observed for the selected
% atom there is no need to consider the remaining part of the query.
% A lot of instance constraints can be pending on this remaining
% part and there are two good reasons for removing those
% constraints from the execution state.
% - efficiency, why spend execution time om constraints that
%   has no effect at all,
% - readable answers, if they are not removed, the answers printed out by
%   the system will contain a lot of irrelevant constraints.
%
% We have added a device for removing such constraints, but using it
% implies that we should be very careful not to remove constraints too early.
% For example, if some subcomputation is still to copy a query together with
% its instance constraints, it would not be smart to remove them!
%
% Our code applies a principle: Always produce copy before splitting the
% computation. Removal of constraints occurs when end-of-program
% is reached (so the copy for this "branch" was produced in vain)
% and when object unification is seen to fail.
%
% This approach is simple and results in reasonably readable code,
% but has the disadvantage that many constraints are copied with the constraints
% immediately being removed.
% Some optimization can be made by adding strange arguments to the
% constraints - but this will destroy the reading of answers
%

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
