%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: es_macros.pl

:- module(es_macros, [translate_macro/2]).
:- use_module(loader).
:- demo_use_module(uti, [append_names/3]).

translate_macro(macro1(Pred_name, Pattern, Condition),
	[(:-block(Block_name)),(:-block(Block_name_aux)),Clause1,Clause2,Clause3,Clause4] ) :-
	atom(Pred_name),
	es:translate(Pattern,TPattern),
	es:translate(Condition,TCondition),
	!,
	Block_name=..[Pred_name,-],

	append_names(Pred_name,'_aux',Pred_name_aux),
	Block_name_aux=..[Pred_name_aux,-],

	Clause1Head=..[Pred_name,\[]],
	Clause1=(Clause1Head :- true),

	Clause2Head=..[Pred_name,\ (?P1 & ?P2)],
	Clause2Body1=..[Pred_name,P1],
	Clause2Body2=..[Pred_name,P2],
	Clause2=(Clause2Head :- Clause2Body1,Clause2Body2),

	Clause3Head=..[Pred_name,\ [?C|?More]],
	Clause3Body1=..[Pred_name_aux,C],
	Clause3Body2=..[Pred_name,More],
	Clause3=(Clause3Head :- Clause3Body1,Clause3Body2),
	
	Clause4Head=..[Pred_name_aux,TPattern],
	Clause4=(Clause4Head :- TCondition).

translate_macro(_, ( :- true ) ) :-
	!, raise_exception(demo(macro)).


/*
Example:

Shape of macro-call:

    easy_mk_bias_predicate( verb_clauses, \verb(?Word, ?ObjCat, ?SubCat),
                            (constant_(Word), category(ObjCat), category(SubCat)))

expands to

    :- block verb_clauses(-), verb_clauses_aux(-).

    verb_clauses( \[]).

    verb_clauses( \ (?A & ?B)):-
        verb_clauses(A), verb_clauses(B).

    verb_clauses( \ [? C | ?More]):-
        verb_clauses_aux(C),
        verb_clauses(More).

    verb_clauses_aux( \verb(?Word, ?ObjCat, ?SubCat) ):- 
        constant_(Word),
        category(ObjCat),  category(SubCat).


I.e., the general pattern is:

    easy_mk_bias_predicate( PRED_NAME, CLAUSE-PATTERN, CONDITION)

and the stuff expands to something of this sort:

    :- block PRED_NAME(-), PRED_NAME_aux(-).

    PRED_NAME( \[]).

    PRED_NAME( \ (?A & ?B)):-
        PRED_NAME(A), PRED_NAME(B).

    PRED_NAME( \ [ ?C | ?More]):-
        PRED_NAME_aux(C),
        PRED_NAME(More).

    PRED_NAME_aux( CLAUSE-PATTERN ):-
       CONDITION.

*/
%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
