%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: es_operators.pl

:- module(es_operators, [oper/5, reserved_symbol/1, object_module_source/2]).

:- use_module(library(lists)).
:- use_module(loader).

:- demo_use_module('uti', [product2list/2,
	is_variable/1, non_variable/1, first_letter/2, lower_case/1]).

:- multifile es:translate/2.

%%% Dynamic predicates used in assertions
:- dynamic sugar_translate/4.
:- dynamic oper/5.
:- dynamic anonymous/1.
:- dynamic object_module_source/2.

%%% The following are some of the keywords for the metalanguage
:- op(600,fx, [ operator ] ).
:- op(550,xf, [ : ] ).
:- op(500, xfx, [ stands ] ).
:- op(499, fx, [ for ] ).
:- op(2, fx , [ ? ] ).

%%% Demo's reserved symbols
reserved_symbol(prolog_atom).
reserved_symbol(prolog_atomic).
reserved_symbol(integer).
reserved_symbol(float).
reserved_symbol(number).
reserved_symbol(identifier).
reserved_symbol(head).
reserved_symbol(compound).
reserved_symbol(term).
reserved_symbol(variable_id).
reserved_symbol(anonymous_variable_id).
has_arguments(head).
has_arguments(compound).
has_arguments(term).

%%% Anonymous variables counter
anonymous(0).

%%%%%%%%%%%%%%%%%%% MODULES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Translation of module declarations
es:translate(object_module(Name,Program), ( :- true ) ) :-
	atom(Name), ground(Program),
	es:translate(Program,Translation), !,
	retractall( object_module_source(Name,_) ),
	assert( object_module_source(Name,Translation) ).

es:translate(object_module(_,_), ( :- true ) ) :-
	!, raise_exception(demo(module)).

%%% Executes the code getting rid of blocked calls
%%% and considers exceptions as failure 
do(Code) :- on_exception(_, call_residue(Code, _), false).

%%%%%%%%%%%%%%%%%%% OPERATORS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
es:translate((operator Left --> Right ), ( :- true) ) :-
	(	Left = ( Operator : InTypes ), intypes(InTypes,ListInTypes) ;
		Left = ( Operator : ), ListInTypes=[] ),
		select_options(Operator,IN,O,Kind),!,
	(	Right = ( T, { Code } stands for OUT ) ;
		Right = ( T stands for OUT ), Code = ( true ) ;
		Right = ( T, { Code } ), OUT = default;
		Right = ( T ), Code = ( true ), OUT = default ),
	es_types:type_exists(T), es:safe_translate(OUT,TOUT),
	asserta( oper(O,ListInTypes,T,((IN,TOUT)) - { Code }, Kind ) ).

intypes(X,_):-var(X),!,fail.
intypes(*(X),*(X)):-!,es_types:type_exists(X).
intypes(+(X),+(X)):-!,es_types:type_exists(X).
intypes(P,L) :- product2list(P,L), es_types:type_exist(L).

es:translate((operator O), ( :- true) ) :- !, raise_exception(demo(operator(O))).


% check_notation(Kind, X, Functor, ParameterizedFunctor) :-


check_notation(R,_,F,ParFunc) :- \+ reserved_symbol(R),!,(F==R -> ParFunc=F).

%% for the following 2 cases we drop the constraint about F being non_variable
check_notation(head, X, F, par(F)) :- X = (F ^ _), is_variable(F),!.
check_notation(head, X, F, F) :- X = (F ^ _),!, (var(F),! ; atom(F) ).
check_notation(compound, X, F, par(F)) :- X = (F ^ L), is_variable(F),!,
	(var(L), ! ; L=[_|_]).
check_notation(compound, X, F, F) :- X = (F ^ L),!, (var(F),! ; atom(F) ),
	(var(L), ! ; L=[_|_]).

check_notation(Kind, X, F, F) :- check_prolog(Kind,X).

%%% other cases of notation2
check_prolog(term,X) :- X = (F ^ _),!, (var(F),! ; atom(F),non_variable(F) ).

%%% Ordinary notation
check_prolog(prolog_atom,X) :- atom(X),non_variable(X).
check_prolog(prolog_atomic,X) :- atomic(X),non_variable(X).
check_prolog(integer,X) :- integer(X).
check_prolog(float,X) :- float(X).
check_prolog(number,X) :- number(X).
check_prolog(identifier,X) :- atom(X),first_letter(X,F),lower_case(F).
check_prolog(head,X) :- compound(X) -> X=..[H|_],non_variable(H),
	dif(H,','),dif(H,';'),dif(H,':-'),dif(H,dif),dif(H,=) ;
	non_variable(X), dif(X,true), \+ number(X).
check_prolog(compound,X) :- compound(X),non_variable(X).
check_prolog(term,_).
check_prolog(variable_id,X) :- is_variable(X),dif(X,'_').
check_prolog(anonymous_variable_id,'_').

select_options(X,_,_,_):-var(X),!,fail.
select_options(any-Options,IN,O,special):-!,parse_options(Options,IN,O).
select_options(O,_,O,ordinary):-atom(O).

parse_options(Options,IN,R) :-
	(
	Options = (R^F), var(F) -> IN=(F^_) ;
	Options = (R^F^Args), has_arguments(R) -> IN=(F^Args) ;
	Options = (R^F), nonvar(F) -> IN=(F^_) ;
	Options = (R) -> IN=_^_ ),
	reserved_symbol(R).

sugar_translate(_,_,X,_):- var(X),!,raise_exception(demo(unexpected_variable)).

sugar_translate(_,_,?X,X):-!.

sugar_translate(TL,Type,Term,Out ) :- 
	oper(R,InTypes,Type,((F^TArgs),O)- { Code },_),
	member(Type,TL),
	( O==default -> Out=gr(Type,ParFunc,TArgs) ; Out = O ),
	( Term=(QF^Args) -> unquote(QF,F),unquote(Args,A),T=F^A ;
	Term=..[F|Args],length(Args,N),	check_special(R, InTypes,N), T=Term ),
	check_notation(R,T,F,ParFunc),
	( do(Code) -> sugar_translate_list(InTypes,Args,TArgs)
	;
	 raise_exception(demo(expression(Term))) ).   

unquote(X,_):- var(X),!,raise_exception(demo(unexpected_variable)).
unquote(?X,X):-!.
unquote(X,X).

check_special(R, _,_):- \+ reserved_symbol(R),!.
check_special(_,List,N):-is_list(List),!,length(List,N).
check_special(_,Special,_):-Special= *(_).
check_special(_,Special,N):-Special= +(_),N >= 1.

get_all_direct_subtypes([],[]):-!.
get_all_direct_subtypes([Type|Tail],Subtypes) :-
	es_types:direct_subtypes(Type,Sub1),
	get_all_direct_subtypes(Tail,Sub2),
	append(Sub1,Sub2,Subtypes).

%%% rule for subtypes
/*
sugar_translate(_,Type,Term,Out) :- 
	var(Type) -> raise_exception(demo(expression(Term))) ;
	es_types:direct_subtypes(Type,SubTypes), sugar_translate(SubTypes,_,Term,Out).
*/
%%%%%%%%%%%%%%%%%%%%%%NEW
sugar_translate(TL,Type,Term,Out) :- 
	ground(Type) -> es_types:direct_subtypes(Type,SubTypes),
		sugar_translate(SubTypes,_,Term,Out) ;
	get_all_direct_subtypes(TL,SUBTL),
	(
	SUBTL==[] -> raise_exception(demo(expression(Term))) ;
	sugar_translate(SUBTL,_,Term,Out)
	).
%%%%%%%%%%%%%%%%%%%%%%END OF NEW

%%% Unquoting operator
sugar_translate_list(_,X,_):- var(X),!,raise_exception(demo(unexpected_variable)).
sugar_translate_list(_,?X,X):-!.

%%% Rules for (*) and (+)
sugar_translate_list(Special,[],[]):-
	(Special= *(_),! ; Special= +(_),! ).
sugar_translate_list(Special,[T|TS],[TT|TTS]):-
	(Special= *(Type),! ; Special= +(Type),! ),
	sugar_translate([Type],Type,T,TT),!,sugar_translate_list(Special,TS,TTS).


%%% Rules for lists
sugar_translate_list([],[],[]).
sugar_translate_list([T|Types],[A|Args],[TA|TArgs]) :-
	sugar_translate([T],T,A,TA),
	sugar_translate_list(Types,Args,TArgs).


%%% Generates a symbol which looks like an anonymous variable
generate_variable_id(V):-retract(anonymous(X)),X1 is X+1,assert(anonymous(X1)),
	anonymize(X1,V).

%%% Puts an underscore in front of a number
anonymize(N,UN) :-
	number_chars(N,Nname),
	append("_", Nname, UNname),
	atom_chars(UN,UNname).

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
