%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: ic_o.pl

:- use_module(library(terms)).
:- use_module(library(lists)).

constraints instance_/3,
            meta_level_quantified/1.



meta_level_quantified(X) <=> ground(X) | true.

meta_level_quantified(X) <=> nonvar(X) | meta_quantify(X).

%%% puts constraints on all variables in X
meta_quantify(X) :- term_variables(X,Vars), meta_quantify_list(Vars).
meta_quantify_list([]):-!.
meta_quantify_list([V1|Tail]):- meta_level_quantified(V1),meta_quantify_list(Tail).

iRule1 @ instance_(gr(T,Sym,XList), Y, S) <=> nonvar(T), T \== variable, \+ subsumes_chk(par(_),Sym) |
	Y = gr(T,Sym,YList),
	(var(Sym) -> meta_level_quantified(Sym) ;true),
	instance_(XList, YList, S).

iRule2 @ instance_(V, X1, S) # Id \ instance_(V, X2, S) <=>
	unify_with_occurs_check(X1,X2) pragma passive(Id).

iRule3a @ instance_(gr(variable,Vname,[]), Value, S) <=> nonvar(Vname) |
	member( (Vname -> Value1), S),
	!, unify_with_occurs_check(Value,Value1).

iRule3b @ instance_(gr(T,par(Id),XList), Value, S) <=> nonvar(T), nonvar(Id) |
	member( (Id -> TId), S),
	!, Value = gr(T,TId,YList), instance_(XList, YList, S).

iRule4 @ instance_(X, gr(T,Sym,YList), S) <=> non_term(T) | 
	X = gr(T,Sym,XList),
	instance_(XList, YList, S).

%%% List of types
iRule5a @ instance_(XL, [], _) <=> XL=[].

iRule5b @ instance_([], YL, _) <=> YL=[].

iRule6a @ instance_(XL, [HY|TY], S) <=>
	XL=[HX|TX],
	instance_(HX,HY,S), instance_(TX,TY,S).

iRule6b @ instance_([HX|TX], YL, S) <=>
	YL=[HY|TY],
	instance_(HX,HY,S), instance_(TX,TY,S).

%%% hand-written declaration of the category of non-terms
non_term(X):- var(X),!,fail.
non_term(formula).
non_term(atom).
non_term(constraint).
non_term(clause).
non_term(program).

instance_(X,Y) :- instance_(X,Y,_).

% Rules for the intersection between instance and type constraints:
% OBS: the two following may threaten safeness!

%%%%%% THE FOLLOWING REFORMULATED DUE TO BORING SYNCHRONIZATION PROBLEM
%%%%%%  - a guard in fail depends on either instance_ or
%%%%%%    meta_level_quantified - so there should not be a short time
%%%%%%    interval where both are not present in the constraint store.
%
%itRule1 @ type(constant,C) \ instance_(C,Y,_) <=>
%	Y=C, meta_level_quantified(C).
%
%irRule2 @ type(structure,X) \ instance_(X,gr(structure,F,AY),S) <=>
%        X = gr(structure,F,AX), instance_(AX,AY,S).


itRule1 @ type(constant,C), instance_(C,Y,_) <=>
	meta_level_quantified(C),
        Y=C,
        type(constant,C).
%%        insert_constraint(type(constant,C),_). % for some reason, this does not work

% PERHAPS THE FOLLOWING MAY GIVE RISE TO SIMILAR PROBLEMS?

irRule2 @ type(structure,X) \ instance_(X,gr(structure,F,AY),S) <=>
        X = gr(structure,F,AX), instance_(AX,AY,S).

/*
instance_(X,Y1,S), X = \\true.
	X = \\true,
	Y1 = \\true
instance_(X,Y,S), Y = \\\a, X = \\\'X'.
	S = [('X'->\\\a)|_A],
	X = \\\'X',
	Y = \\\a
instance_(X,Y,S), Y = \\\f(?A,?B).
	Y = \\\f(?A,?B),
	instance_(X,\\\f(?A,?B),S)
instance_(X,Y,S), X = \\\f(?A,?B).
	X = \\\f(?A,?B),
	Y = \\\f(?_A,?_B),
	instance_(B,_B,S),
	instance_(A,_A,S)
instance_(X,\\\ f(a),S1),instance_(X,\\\ f(b),S2),X= \\\f(?Z).
	X = \\\f(?Z),
	instance_(Z,\\\b,S2),
	instance_(Z,\\\a,S1)

*/

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
