%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: oc.pl

constraints ground_/1, not_member_/2.

%%%%%%%%%%
% GROUND %
%%%%%%%%%%

ground_(gr(Type,_,AList)) <=> true |
    Type == variable -> fail
    ;
    ground_(AList).

% where

ground_([]) <=> true.


ground_([A|As]) <=> ground_(A), ground_(As).

ground_(X), type(variable,X) <=> fail.

% OBS: the two following may threaten safeness!
ground_(X), instance_(X, Y, _) <=> X=Y, meta_level_quantified(X).

ground_(Y), instance_(X, Y, _) <=> X=Y, meta_level_quantified(X).

%%%%%%%%%%
% MEMBER %
%%%%%%%%%%

% Defined in straightforward Sicstus Prolog
% (M1)
member_(C, V) :- var(V), !, V = \[?C | ?_].

% (M2)
member_(C, \[?C | ?_]).

% (M3)
member_(C, \[?_ | ?P1]) :- member_(C, P1).

% (M&)
member_(C,  \(?P1 & ?P2)) :- member_(C, P1) ; member_(C, P2).


%%%%%%%%%%%%%%%%%
% NO DUPLICATES %
%%%%%%%%%%%%%%%%%

:- block no_duplicates_(-).

% (ND1)
no_duplicates_(\ [?C|?P]) :-
	not_member_(C,P),
	no_duplicates_(P).

% (ND2)
no_duplicates_(\[]).

no_duplicates_( \ ( ?A & ?B) ) :-
	no_duplicates_(A),
	no_duplicates_(B),
	disjoint_(A,B).

%%%% DISTRIBUTION RULE OVER &
:- block disjoint_(-,-).

disjoint_(A,B):-
	nonvar(A) ->
		(A = \ (?A1 & ?A2) -> disjoint_(A1,B), disjoint_(A2,B), disjoint_(A1,A2) ;
        A = \ [?C|?Cs] -> not_member_(C, B), disjoint_(Cs,B) ;
        A = \[] -> true)
	; % nonvar(B)
		(B = \ (?B1 & ?B2) -> disjoint_(A,B1), disjoint_(A,B2), disjoint_(B1,B2) ;
		B = \ [?C|?Cs] -> not_member_(C, A), disjoint_(Cs,A) ;
		B = \[] -> true).


%%%%%%%%%%%%%%
% NOT MEMBER %
%%%%%%%%%%%%%%


/* NEW DEFINITION as constraint CLAUSES TAKING INTO ACCOUNT preds_ and no_preds_ */
preds_(P,Spec) \ not_member_(\ (?Pred ^ ?_ :- ?_), P) <=> \+ member(Pred,Spec) | true.

%%%% not sure about the following
no_preds_(P,Spec) \ not_member_(\ (?Pred ^ ?_ :- ?_), P) <=>  member(Pred,Spec) | true.

%% the remaining are just a rewriting of the Prolog code
not_member_(C, \ [?C1|?P]) <=>
    dif(C,C1),
    not_member_(C, P).

not_member_(_, \[]) <=> true.

not_member_(C, \ ( ?P1 & ?P2 ) ) <=>
    not_member_(C, P1), not_member_(C, P2).
/* END OF NEW CLAUSES */

/*
:- block not_member_(?, -).

not_member_(C, \ [?C1|?P]):-
    dif(C,C1),
    not_member_(C, P).

not_member_(_, \ []).

not_member_(C, \ ( ?P1 & ?P2 ) ):-
    not_member_(C, P1), not_member_(C, P2).
*/
%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
