%%%%%%% This file is part of the DemoII system produced
%%%%%%% 1997-1999 by Davide Martinenghi and Henning Christiansen
%%%%%%%
%%%%%%% Contact:    henning@ruc.dk
%%%%%%%
%%%%%%% See legal notice in the end of this file.
%%%%%%%
%%%%%%% The code has been developed and tested for SICStus Prolog 3.7
%%%%%%%
%%%%%%% Filename: sc.pl

:- demo_use_module('Source/uti').

constraints prolog_check/2.


%% the following 3 are written because we actually refer only
%% to the first symbol of the head when we call the constraint
prolog_check(head,X) <=> prolog_check(function_id_or_parameter,X).
prolog_check(compound,X) <=> prolog_check(function_id_or_parameter,X).
prolog_check(term,X) <=> prolog_check(function_id_or_parameter,X).

prolog_check(function_id_or_parameter,par(P)) <=> prolog_check(variable_id,P).
prolog_check(function_id_or_parameter,Y) <=> nonvar(Y) | prolog_check(function_id,Y).

%%% rule for intersection of type- and syntax- constraints
prolog_check(_,X), type(_,X) <=> fail.

%%% rule for categories in a subset relation
prolog_check(Kind1,Y),prolog_check(Kind2,Y) # Id <=> var(Y) |
	(
	syn_subset(Kind1,Kind2) -> prolog_check(Kind1,Y) ;
	syn_subset(Kind2,Kind1) -> prolog_check(Kind2,Y) ;
	fail	%%% the categories are disjoint
	) pragma passive(Id).

%%% normal rule
prolog_check(X,Y) <=> nonvar(X),nonvar(Y) | check2(X,Y).

check2(prolog_atomic,X) :- atomic(X),non_variable(X).
check2(integer,X) :- integer(X).
check2(float,X) :- float(X).
check2(number,X) :- number(X).
check2(identifier,X) :- first_letter(X,F),lower_case(F).
check2(function_id,X) :- atom(X), non_variable(X),
	dif(X,','),dif(X,';'),dif(X,':-'),dif(X,dif),dif(X,=), dif(X,true).
check2(variable_id,X) :- is_variable(X),
	\+ prolog_check(anonymous_variable_id,X).
check2(anonymous_variable_id,X) :- name(X,[95|T]), name(N,T), number(N).

%%% Syntactical categories
direct_syn_subset(function_id,function_id_or_parameter).
direct_syn_subset(parameter,function_id_or_parameter).	% this is not used
direct_syn_subset(function_id_or_parameter,prolog_atomic).
direct_syn_subset(variable_id,prolog_atomic).
direct_syn_subset(anonymous_variable_id,prolog_atomic).
direct_syn_subset(number,prolog_atomic).
direct_syn_subset(integer,number).
direct_syn_subset(float,number).

syn_subset(X,Y) :- direct_syn_subset(X,Y).
syn_subset(X,Y) :- direct_syn_subset(X,Z),syn_subset(Z,Y).

operator_exists(Sym, InTypes, OutType, List) :-
	user:oper(Sym,InTypes,OutType,(_,default)-{true},ordinary) ;
	user:oper(R,InTypes,OutType,((Sym^List),_)- { _ },special), prolog_check(R,Sym).

%%%%%%%
%%%%%%% Copyright (C) Henning Christiansen 1998, Roskilde University, Denmark
%%%%%%% E-mail henning@ruc.dk
%%%%%%% 
%%%%%%% The code contained in this file must not be applied
%%%%%%% in any commercial product without written permission
%%%%%%% from the copyright holder.
%%%%%%%
%%%%%%% Permission is granted to use, copy, modify, and distribute
%%%%%%% this file for scientific and educational purposes provided
%%%%%%% that the copyright notice and these remarks are maintained.
%%%%%%% The copyright holder or author takes no responsibility for
%%%%%%% any loss or damage caused by this code.
%%%%%%%
%%%%%%% Web site for the DemoII system: http://www.dat.ruc.dk/software/index.html
